#!/usr/bin/env python3

"""Run AFL repeatedly with externally supplied generated packet from STDIN."""

from collections import namedtuple

import afl  # pylint: disable=import-error
import sys
from chewie.mac_address import MacAddress
from chewie.message_parser import MessageParser
from chewie.utils import MessageParseError

ROUNDS = 1


class NoneDict(dict):
    """Dictionary that will always return None"""

    def __getitem__(self, key):
        return None  # pylint: disable=useless-return


def main(eap):
    """Run AFL repeatedly with externally supplied generated packet from STDIN."""

    while afl.loop(ROUNDS):
        # receive input from afl
        rcv = sys.stdin.read()
        data = None
        try:
            data = bytearray.fromhex(rcv)  # pytype: disable=missing-parameter
        except (ValueError, TypeError):
            return
        if eap:
            test_one_x_parse(data)
        else:
            test_radius_parse(data)


def test_one_x_parse(data):
    """Tests the one_x_parse function
    Args:
        data: payload to parse"""
    try:
        MessageParser.one_x_parse(data, MacAddress.from_string("00:00:00:12:34:56"))
    except MessageParseError:
        # Ignore exceptions the parser intentionally throws, and are caught by the caller.
        pass


def test_radius_parse(data):
    """Tests the radiuse_parse function
    Args:
        data: payload to parse"""

    try:
        # The dict sets the packet ID for the known packet_id (test/fuzzer/radius_packet*.ex)
        #  to none. so no validation is done.
        MessageParser.radius_parse(data, "SECRET",
                                   radius_lifecycle=namedtuple('RadiusLifecycle',
                                                               'packet_id_to_request_authenticator')
                                   (NoneDict()))
    except MessageParseError:
        # Ignore exceptions the parser intentionally throws, and are caught by the caller.
        pass


if __name__ == "__main__":
    PARSER = sys.argv[1]
    EAP = False
    if PARSER == 'eap':
        EAP = True
    main(EAP)



# pylint: disable=missing-docstring
import struct
import unittest

from chewie.eap import Eap
from chewie.mac_address import MacAddress
from chewie.message_parser import EapolStartMessage, EapolLogoffMessage, \
    SuccessMessage, FailureMessage
from chewie.message_parser import MessageParser, MessagePacker, IdentityMessage, \
    Md5ChallengeMessage, TtlsMessage, \
    LegacyNakMessage, TlsMessage, PeapMessage
from chewie.radius_attributes import State, CalledStationId, NASPortType
from chewie.utils import MessageParseError


class MessageParserTestCase(unittest.TestCase):
    def test_identity_request_message_parses(self):  # pylint: disable=invalid-name
        packed_message = bytes.fromhex("0180c2000003001906eab88c888e010000050101000501000000")
        message = MessageParser.ethernet_parse(packed_message)[0]
        self.assertEqual(MacAddress.from_string("00:19:06:ea:b8:8c"), message.src_mac)
        self.assertEqual(1, message.message_id)
        self.assertEqual(Eap.REQUEST, message.code)
        self.assertEqual("", message.identity)

    def test_identity_response_message_parses(self):  # pylint: disable=invalid-name
        packed_message = bytes.fromhex(
            "0180c2000003001422e9545e888e0100001102000011014a6f686e2e4d63477569726b")
        message = MessageParser.ethernet_parse(packed_message)[0]
        self.assertEqual(MacAddress.from_string("00:14:22:e9:54:5e"), message.src_mac)
        self.assertEqual(0, message.message_id)
        self.assertEqual(Eap.RESPONSE, message.code)
        self.assertEqual("John.McGuirk", message.identity)

    def test_md5_challenge_request_message_parses(self):  # pylint: disable=invalid-name
        packed_message = bytes.fromhex(
            "0180c2000003001906eab88c888e01000016010100160410824788d693e2adac6ce15641418228cf0000")
        message = MessageParser.ethernet_parse(packed_message)[0]
        self.assertEqual(MacAddress.from_string("00:19:06:ea:b8:8c"), message.src_mac)
        self.assertEqual(1, message.message_id)
        self.assertEqual(Eap.REQUEST, message.code)
        self.assertEqual(bytes.fromhex("824788d693e2adac6ce15641418228cf"), message.challenge)
        self.assertEqual(b"", message.extra_data)

    def test_md5_challenge_response_message_parses(self):  # pylint: disable=invalid-name
        packed_message = bytes.fromhex(
            "0180c2000003001422e9545e888e010000220201002204103a535f0ee8c6b34fe714aa7dad9a0e154a6f686e2e4d63477569726b")  # pylint: disable=line-too-long
        message = MessageParser.ethernet_parse(packed_message)[0]
        self.assertEqual(MacAddress.from_string("00:14:22:e9:54:5e"), message.src_mac)
        self.assertEqual(1, message.message_id)
        self.assertEqual(Eap.RESPONSE, message.code)
        self.assertEqual(bytes.fromhex("3a535f0ee8c6b34fe714aa7dad9a0e15"), message.challenge)
        self.assertEqual(b"John.McGuirk", message.extra_data)

    def test_identity_request_message_packs(self):  # pylint: disable=invalid-name
        expected_packed_message = bytes.fromhex(
            "0180c2000003001906eab88c888e010000050101000501")
        message = IdentityMessage(src_mac=MacAddress.from_string("00:19:06:ea:b8:8c"), message_id=1,
                                  code=Eap.REQUEST, identity="")
        packed_message = MessagePacker.ethernet_pack(message,
                                                     MacAddress.from_string("00:19:06:ea:b8:8c"),
                                                     MacAddress.from_string("01:80:c2:00:00:03"))
        self.assertEqual(expected_packed_message, packed_message)

    def test_identity_response_message_packs(self):  # pylint: disable=invalid-name
        expected_packed_message = bytes.fromhex(
            "0180c2000003001422e9545e888e0100001102000011014a6f686e2e4d63477569726b")
        message = IdentityMessage(src_mac=MacAddress.from_string("00:14:22:e9:54:5e"),
                                  message_id=0, code=Eap.RESPONSE, identity="John.McGuirk")
        packed_message = MessagePacker.ethernet_pack(message,
                                                     MacAddress.from_string("00:14:22:e9:54:5e"),
                                                     MacAddress.from_string("01:80:c2:00:00:03"))
        self.assertEqual(expected_packed_message, packed_message)

    def test_md5_challenge_request_message_packs(self):  # pylint: disable=invalid-name
        expected_packed_message = bytes.fromhex(
            "0180c2000003001906eab88c888e01000016010100160410824788d693e2adac6ce15641418228cf")
        message = Md5ChallengeMessage(src_mac=MacAddress.from_string("00:19:06:ea:b8:8c"),
                                      message_id=1,
                                      code=Eap.REQUEST,
                                      challenge=bytes.fromhex(
                                          "824788d693e2adac6ce15641418228cf"),
                                      extra_data=b"")
        packed_message = MessagePacker.ethernet_pack(message,
                                                     MacAddress.from_string("00:19:06:ea:b8:8c"),
                                                     MacAddress.from_string("01:80:c2:00:00:03"))
        self.assertEqual(expected_packed_message, packed_message)

    def test_md5_challenge_response_message_packs(self):  # pylint: disable=invalid-name
        expected_packed_message = bytes.fromhex(
            "0180c2000003001422e9545e888e010000220201002204103a535f0ee8c6b34fe714aa7dad9a0e154a6f686e2e4d63477569726b")  # pylint: disable=line-too-long
        message = Md5ChallengeMessage(src_mac=MacAddress.from_string("00:14:22:e9:54:5e"),
                                      message_id=1,
                                      code=Eap.RESPONSE,
                                      challenge=bytes.fromhex(
                                          "3a535f0ee8c6b34fe714aa7dad9a0e15"),
                                      extra_data=b"John.McGuirk")
        packed_message = MessagePacker.ethernet_pack(message,
                                                     MacAddress.from_string("00:14:22:e9:54:5e"),
                                                     MacAddress.from_string("01:80:c2:00:00:03"))
        self.assertEqual(expected_packed_message, packed_message)

    def test_eapol_start_message_parses(self):
        packed_message = bytes.fromhex("0180c2000003001906eab88c888e01010000")
        message = MessageParser.ethernet_parse(packed_message)[0]
        self.assertEqual(MacAddress.from_string("00:19:06:ea:b8:8c"), message.src_mac)
        self.assertTrue(isinstance(message, EapolStartMessage))

    def test_eapol_start_message_packs(self):
        expected_packed_message = bytes.fromhex("0180c2000003001906eab88c888e01010000")
        message = EapolStartMessage(src_mac=MacAddress.from_string("00:19:06:ea:b8:8c"))
        packed_message = MessagePacker.ethernet_pack(message,
                                                     MacAddress.from_string("00:19:06:ea:b8:8c"),
                                                     MacAddress.from_string("01:80:c2:00:00:03"))
        self.assertEqual(expected_packed_message, packed_message)

    def test_eapol_logoff_message_parses(self):  # pylint: disable=invalid-name
        packed_message = bytes.fromhex("0180c2000003001906eab88c888e01020000")
        message = MessageParser.ethernet_parse(packed_message)[0]
        self.assertEqual(MacAddress.from_string("00:19:06:ea:b8:8c"), message.src_mac)
        self.assertTrue(isinstance(message, EapolLogoffMessage))

    def test_eapol_logoff_message_packs(self):
        expected_packed_message = bytes.fromhex("0180c2000003001906eab88c888e01020000")
        message = EapolLogoffMessage(src_mac=MacAddress.from_string("00:19:06:ea:b8:8c"))
        packed_message = MessagePacker.ethernet_pack(message,
                                                     MacAddress.from_string("00:19:06:ea:b8:8c"),
                                                     MacAddress.from_string("01:80:c2:00:00:03"))
        self.assertEqual(expected_packed_message, packed_message)

    def test_success_message_parses(self):
        packed_message = bytes.fromhex("0180c2000003001906eab88c888e0100000403ff0004")
        message = MessageParser.ethernet_parse(packed_message)[0]
        self.assertEqual(MacAddress.from_string("00:19:06:ea:b8:8c"), message.src_mac)
        self.assertEqual(255, message.message_id)
        self.assertTrue(isinstance(message, SuccessMessage))

    def test_success_message_packs(self):
        expected_packed_message = bytes.fromhex("0180c2000003001906eab88c888e0100000403ff0004")
        message = SuccessMessage(src_mac=MacAddress.from_string("00:19:06:ea:b8:8c"),
                                 message_id=255)
        packed_message = MessagePacker.ethernet_pack(message,
                                                     MacAddress.from_string("00:19:06:ea:b8:8c"),
                                                     MacAddress.from_string("01:80:c2:00:00:03"))
        self.assertEqual(expected_packed_message, packed_message)

    def test_failure_message_parses(self):
        packed_message = bytes.fromhex("0180c2000003001906eab88c888e0100000404ff0004")
        message = MessageParser.ethernet_parse(packed_message)[0]
        self.assertEqual(MacAddress.from_string("00:19:06:ea:b8:8c"), message.src_mac)
        self.assertEqual(255, message.message_id)
        self.assertTrue(isinstance(message, FailureMessage))

    def test_failure_message_packs(self):
        expected_packed_message = bytes.fromhex("000000000001001906eab88c888e0100000404ff0004")
        message = FailureMessage(src_mac=MacAddress.from_string("00:19:06:ea:b8:8c"),
                                 message_id=255)
        packed_message = MessagePacker.ethernet_pack(message,
                                                     MacAddress.from_string("00:19:06:ea:b8:8c"),
                                                     MacAddress.from_string("00:00:00:00:00:01"))
        self.assertEqual(expected_packed_message, packed_message)

    def test_ttls_message_parses(self):
        packed_message = bytes.fromhex("000000111101444444444444888e"
                                       "01000006016900061520")
        message = MessageParser.ethernet_parse(packed_message)[0]
        self.assertEqual(MacAddress.from_string("44:44:44:44:44:44"), message.src_mac)
        self.assertEqual(105, message.message_id)
        self.assertIsInstance(message, TtlsMessage)

    def test_ttls_message_packs(self):
        expected_packed_message = bytes.fromhex("000000111101444444444444888e"
                                                "01000006016900061520")
        message = TtlsMessage(src_mac=MacAddress.from_string("44:44:44:44:44:44"),
                              message_id=105, code=Eap.REQUEST,
                              flags=0x20, extra_data=b'')
        packed_message = MessagePacker.ethernet_pack(message,
                                                     MacAddress.from_string("44:44:44:44:44:44"),
                                                     MacAddress.from_string("00:00:00:11:11:01"))
        self.assertEqual(expected_packed_message, packed_message)

    def test_tls_message_parses(self):
        packed_message = bytes.fromhex("000000111101444444444444888e"
                                       "010000b2026900b20d0016030100a7010000a303038c8007fa4ffe8f11fbe62debce4a1385e70be51efe77b105d205d2dc9ae815a5000038c02cc030009fcca9cca8ccaac02bc02f009ec024c028006bc023c0270067c00ac0140039c009c0130033009d009c003d003c0035002f00ff01000042000b000403000102000a000a0008001d0017001900180016000000170000000d0020001e060106020603050105020503040104020403030103020303020102020203")  # pylint: disable=line-too-long
        message = MessageParser.ethernet_parse(packed_message)[0]
        self.assertEqual(MacAddress.from_string("44:44:44:44:44:44"), message.src_mac)
        self.assertEqual(105, message.message_id)
        self.assertEqual(0, message.flags)
        self.assertIsInstance(message, TlsMessage)

    def test_tls_message_packs(self):
        expected_packed_message = bytes.fromhex("000000111101444444444444888e"
                                                "010000b2026900b20d0016030100a7010000a303038c8007fa4ffe8f11fbe62debce4a1385e70be51efe77b105d205d2dc9ae815a5000038c02cc030009fcca9cca8ccaac02bc02f009ec024c028006bc023c0270067c00ac0140039c009c0130033009d009c003d003c0035002f00ff01000042000b000403000102000a000a0008001d0017001900180016000000170000000d0020001e060106020603050105020503040104020403030103020303020102020203")  # pylint: disable=line-too-long
        message = TlsMessage(src_mac=MacAddress.from_string("44:44:44:44:44:44"),
                             message_id=105, code=Eap.RESPONSE,
                             flags=0x00, extra_data=bytes.fromhex(
                '16030100a7010000a303038c8007fa4ffe8f11fbe62debce4a1385e70be51efe77b105d205d2dc9ae815a5000038c02cc030009fcca9cca8ccaac02bc02f009ec024c028006bc023c0270067c00ac0140039c009c0130033009d009c003d003c0035002f00ff01000042000b000403000102000a000a0008001d0017001900180016000000170000000d0020001e060106020603050105020503040104020403030103020303020102020203'))  # pylint: disable=line-too-long
        packed_message = MessagePacker.ethernet_pack(message,
                                                     MacAddress.from_string("44:44:44:44:44:44"),
                                                     MacAddress.from_string("00:00:00:11:11:01"))
        self.assertEqual(expected_packed_message, packed_message)

    def test_peap_message_parses(self):
        packed_message = bytes.fromhex("000000111101444444444444888e"
                                       "010000b2026900b2190016030100a7010000a303038c8007fa4ffe8f11fbe62debce4a1385e70be51efe77b105d205d2dc9ae815a5000038c02cc030009fcca9cca8ccaac02bc02f009ec024c028006bc023c0270067c00ac0140039c009c0130033009d009c003d003c0035002f00ff01000042000b000403000102000a000a0008001d0017001900180016000000170000000d0020001e060106020603050105020503040104020403030103020303020102020203")  # pylint: disable=line-too-long
        message = MessageParser.ethernet_parse(packed_message)[0]
        self.assertEqual(MacAddress.from_string("44:44:44:44:44:44"), message.src_mac)
        self.assertEqual(105, message.message_id)
        self.assertEqual(0, message.flags)
        self.assertIsInstance(message, PeapMessage)

    def test_peap_message_packs(self):
        expected_packed_message = bytes.fromhex("000000111101444444444444888e"
                                                "010000b2026900b2190016030100a7010000a303038c8007fa4ffe8f11fbe62debce4a1385e70be51efe77b105d205d2dc9ae815a5000038c02cc030009fcca9cca8ccaac02bc02f009ec024c028006bc023c0270067c00ac0140039c009c0130033009d009c003d003c0035002f00ff01000042000b000403000102000a000a0008001d0017001900180016000000170000000d0020001e060106020603050105020503040104020403030103020303020102020203")  # pylint: disable=line-too-long
        message = PeapMessage(src_mac=MacAddress.from_string("44:44:44:44:44:44"),
                              message_id=105, code=Eap.RESPONSE,
                              flags=0x00,
                              extra_data=bytes.fromhex(
                                  '16030100a7010000a303038c8007fa4ffe8f11fbe62debce4a1385e70be51efe77b105d205d2dc9ae815a5000038c02cc030009fcca9cca8ccaac02bc02f009ec024c028006bc023c0270067c00ac0140039c009c0130033009d009c003d003c0035002f00ff01000042000b000403000102000a000a0008001d0017001900180016000000170000000d0020001e060106020603050105020503040104020403030103020303020102020203'))  # pylint: disable=line-too-long
        packed_message = MessagePacker.ethernet_pack(message,
                                                     MacAddress.from_string("44:44:44:44:44:44"),
                                                     MacAddress.from_string("00:00:00:11:11:01"))
        self.assertEqual(expected_packed_message, packed_message)

    def test_legacy_nak_message_parses(self):
        packed_message = bytes.fromhex("0180c2000003000000111101888e"
                                       "01000006026800060315")
        message = MessageParser.ethernet_parse(packed_message)[0]
        self.assertEqual(MacAddress.from_string("00:00:00:11:11:01"), message.src_mac)
        self.assertEqual(104, message.message_id)
        self.assertIsInstance(message, LegacyNakMessage)

    def test_legacy_nak_message_packs(self):
        expected_packed_message = bytes.fromhex("0180c2000003000000111101888e"
                                                "01000006026800060315")
        message = LegacyNakMessage(src_mac=MacAddress.from_string("00:00:00:11:11:01"),
                                   message_id=104,
                                   code=Eap.RESPONSE,
                                   desired_auth_types=[(21).to_bytes(length=1, byteorder='big')])
        packed_message = MessagePacker.ethernet_pack(message,
                                                     MacAddress.from_string("00:00:00:11:11:01"),
                                                     MacAddress.from_string("01:80:c2:00:00:03"))
        self.assertEqual(expected_packed_message, packed_message)

    def test_radius_with_extra_attributes_packs(self):  # pylint: disable=invalid-name

        packed_message = bytes.fromhex("010a0073"
                                       "be5df1f3b3366c69b977e56a7da47cba"
                                       "010675736572"
                                       "1f1330323a34323a61633a31373a30303a3666"
                                       "1e1434342d34342d34342d34342d34342d34343a"
                                       "3d060000000f"
                                       "4f08027100061500"
                                       "1812f51d90b0f76c85835ed4ac882e522748501201531ea8051d136941fece17473f6b4a")  # pylint: disable=line-too-long

        src_mac = MacAddress.from_string("02:42:ac:17:00:6f")
        username = "user"
        radius_packet_id = 10
        request_authenticator = bytes.fromhex("be5df1f3b3366c69b977e56a7da47cba")
        state = State.create(bytes.fromhex("f51d90b0f76c85835ed4ac882e522748"))
        secret = "SECRET"
        extra_attributes = []
        extra_attributes.append(CalledStationId.create('44-44-44-44-44-44:'))
        extra_attributes.append(NASPortType.create(15))

        eap_message = TtlsMessage(src_mac, 113, Eap.RESPONSE, 0, b'')

        packed_radius = MessagePacker.radius_pack(eap_message, src_mac, username, radius_packet_id,
                                                  request_authenticator, state, secret,
                                                  extra_attributes=extra_attributes)

        self.assertEqual(packed_message, packed_radius)

    def test_radius_packs_with_nas_port(self):

        packed_message = bytes.fromhex(
            "01bf00610123456789abcdeffedcba9876543210010a62656e62757274741f1361613a62623a63633a64643a65653a66660506000002a14f18021500160410824788d693e2adac6ce15641418228cf50121139bd192c46fe6d2a937d9573311b70")  # pylint: disable=line-too-long

        src_mac = MacAddress.from_string("aa:bb:cc:dd:ee:ff")
        username = "benburtt"
        radius_packet_id = 191
        request_authenticator = bytes.fromhex("0123456789abcdeffedcba9876543210")
        state = None
        secret = "SUPERSECRET"
        challenge = bytes.fromhex("824788d693e2adac6ce15641418228cf")
        eap_message = Md5ChallengeMessage(src_mac, 21, Eap.RESPONSE, challenge, b'')
        packed_radius = MessagePacker.radius_pack(eap_message, src_mac, username, radius_packet_id,
                                                  request_authenticator, state, secret,
                                                  nas_port=0x02a1)
        self.assertEqual(packed_message, packed_radius)

    def test_radius_packs_basic(self):
        """without extra_attributes or nas-port"""

        packed_message = bytes.fromhex(
            "01bf005b0123456789abcdeffedcba9876543210010a62656e62757274741f1361613a62623a63633a64643a65653a66664f18021500160410824788d693e2adac6ce15641418228cf5012caadc1c7a3be07fe63fdf83a59ed18c2")  # pylint: disable=line-too-long

        src_mac = MacAddress.from_string("aa:bb:cc:dd:ee:ff")
        username = "benburtt"
        radius_packet_id = 191
        request_authenticator = bytes.fromhex("0123456789abcdeffedcba9876543210")
        state = None
        secret = "SUPERSECRET"
        challenge = bytes.fromhex("824788d693e2adac6ce15641418228cf")
        eap_message = Md5ChallengeMessage(src_mac, 21, Eap.RESPONSE, challenge, b'')
        packed_radius = MessagePacker.radius_pack(eap_message, src_mac, username, radius_packet_id,
                                                  request_authenticator, state, secret)
        self.assertEqual(packed_message, packed_radius)

    def test_unicode_decode_error_converts_to_message_parse_error(
            self):  # pylint: disable=invalid-name
        data = bytes.fromhex("0163bf130103bf1301")
        try:
            MessageParser.eap_parse(data, MacAddress.from_string("00:00:00:12:34:56"))
        except MessageParseError as exception:
            self.assertIsInstance(exception.__cause__, UnicodeDecodeError)
            return
        self.fail()

    def test_struct_unpack_error_converts_to_message_parse_error(
            self):  # pylint: disable=invalid-name
        data = bytes.fromhex("01001000")
        try:
            MessageParser.eap_parse(data, MacAddress.from_string("00:00:00:12:34:56"))
        except MessageParseError as exception:
            self.assertIsInstance(exception.__cause__, struct.error)
            return
        self.fail()

    def test_bad_eap_code(self):
        data = bytes.fromhex("001F1101")
        self.assertRaises(MessageParseError,
                          MessageParser.eap_parse,
                          data, MacAddress.from_string("00:00:00:12:34:56"))

    def test_bad_packet_type(self):
        data = bytes.fromhex("02a02a02070207")
        self.assertRaises(MessageParseError,
                          MessageParser.eap_parse,
                          data, MacAddress.from_string("00:00:00:12:34:56"))

    # TODO
    # def test_radius_mab_packs_basic(self):
    #     """without extra_attributes or nas-port"""
    #
    #     packed_message = bytes.fromhex("01630047c7835a254dc85f1e82ec0702956c6d09010e61616262636364646565666602128502bedd390c0020bf4a13966324d99c1e1361612d62622d63632d64642d65652d6666")  # pylint: disable=line-too-long
    #
    #     src_mac = MacAddress.from_string("aa:bb:cc:dd:ee:ff")
    #     radius_packet_id = 99
    #     request_authenticator = bytes.fromhex("c7835a254dc85f1e82ec0702956c6d09")
    #     state = None
    #     secret = "SECRET"
    #
    #     packed_radius = MessagePacker.radius_mab_pack(src_mac, radius_packet_id, request_authenticator, secret, None)
    #     self.assertEqual(packed_message, packed_radius,
    #                      "Did not match\nActual: {}\nExpected: {}".format(
    #                          binascii.hexlify(packed_radius), binascii.hexlify(packed_message)))



"""Loosely based on RFC4137 'EAP State Machines' with some interpretation"""
import random

from transitions import Machine, State

from chewie.eap import Eap
from chewie.event import EventMessageReceived, EventRadiusMessageReceived, EventTimerExpired, \
    EventPortStatusChange, EventSessionTimeout, EventPreemptiveEAPResponseMessageReceived
from chewie.message_parser import SuccessMessage, FailureMessage, EapolStartMessage, \
    IdentityMessage, EapolLogoffMessage, EapMessage

# To create a difference between transitions.State and messageparser.EapMessage
import chewie.radius_attributes as radius_attributes
from chewie.utils import get_logger, log_method, RadiusQueueMessage, EapQueueMessage
from chewie.radius import RadiusPacket

class Policy:
    """Fleshed out enough to support passthrough mode."""

    @staticmethod
    def get_next_method(eap_resp_data):
        # TODO Probably should do something else
        if isinstance(eap_resp_data, EapolStartMessage):
            return "IDENTITY"
        return "IDENTITY"
        # return "NOTIFICATION"

    @staticmethod
    def get_decision(eap_resp_data):
        # TODO if not offloading return success/failure/Continue
        if eap_resp_data is None or isinstance(eap_resp_data, EapolStartMessage):
            return Decision.CONTINUE
        return Decision.PASSTHROUGH

    @staticmethod
    def update():
        # TODO actually do something?
        pass


class MethodState:
    # pylint: disable=too-few-public-methods
    CONTINUE = "CONTINUE"
    END = "END"
    PROPOSED = "PROPOSED"
    IDENTITY = "IDENTITY"
    NOTIFICATION = "NOTIFICATION"
    NAK = "NAK"
    EXPANDED_NAK = "EXPANDED_NAK"


class Decision:
    # pylint: disable=too-few-public-methods
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"
    CONTINUE = "CONTINUE"
    PASSTHROUGH = "PASSTHROUGH"


class MPassthrough:
    """M = Method, so if we wanted to do MD5 locally (not passthrough), we'd
    Have class MMD5 with it's own implementation of the methods below"""

    done = False
    src_mac = None

    def check(self, eap_resp_data):  # pylint: disable=unused-argument
        """
        Args:
             eap_resp_data (Message):
        Returns:
            True if packet should be ignored. otherwise False if packet is good.
        """
        # TODO check the *integrity* of the packet.
        #  The IDs already match (done on entry to INTEGRITY_CHECK state)
        return False

    def process(self, eap_resp_data):
        if isinstance(eap_resp_data, IdentityMessage):
            self.done = True

    def init(self, src_mac):
        self.src_mac = src_mac

    def reset(self):
        self.done = False

    def is_done(self):
        return self.done

    def get_timeout(self):
        return 1

    def get_key(self):
        return None

    def build_req(self, current_id):
        return IdentityMessage(self.src_mac, current_id, Eap.REQUEST, "")


class FullEAPStateMachine:
    """Based on RFC 4137 section 7 (EAP Full Authenticator).
    Only acts in passthrough mode (no local method support).
    """

    # non RFC 4137 variables/CONSTANTs
    DEFAULT_TIMEOUT = 5  # Number of Seconds
    DEFAULT_SESSION_TIMEOUT = 3600  # Number of Seconds

    RADIUS_RETRANSMIT_TIMEOUT = 5

    state = None
    eap_output_messages = None
    src_mac = None
    # TODO can use dst_mac to verify where packet came from more thoroughly
    port_id_mac = None
    radius_state_attribute = None  # the last state from radius server
    sent_count = 0
    session_timeout_job = None

    session_timeout = DEFAULT_SESSION_TIMEOUT
    radius_tunnel_private_group_id = None
    filter_id = None

    machine = None

    NO_STATE = "NO_STATE"
    DISABLED = "DISABLED"
    INITIALIZE = "INITIALIZE"
    IDLE = "IDLE"
    RECEIVED = "RECEIVED"
    INTEGRITY_CHECK = "INTEGRITY_CHECK"
    METHOD_RESPONSE = "METHOD_RESPONSE"
    METHOD_REQUEST = "METHOD_REQUEST"
    PROPOSE_METHOD = "PROPOSED_METHOD"
    SELECT_ACTION = "SELECT_ACTION"
    SEND_REQUEST = "SEND_REQUEST"
    DISCARD = "DISCARD"
    NAK = "NAK"
    RETRANSMIT = "RETRANSMIT"
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"
    TIMEOUT_FAILURE = "TIMEOUT_FAILURE"

    INITIALIZE_PASSTRHOUGH = "INITIALIZE_PASSTHROUGH"
    IDLE2 = "IDLE2"
    RECEIVED2 = "RECEIVED2"
    AAA_REQUEST = "AAA_REQUEST"
    AAA_IDLE = "AAA_IDLE"
    AAA_RESPONSE = "AAA_RESPONSE"
    SEND_REQUEST2 = "SEND_REQUEST2"
    DISCARD2 = "DISCARD2"
    RETRANSMIT2 = "RETRANSMIT2"
    SUCCESS2 = "SUCCESS2"
    FAILURE2 = "FAILURE2"
    TIMEOUT_FAILURE2 = "TIMEOUT_FAILURE2"

    # Non RFC 4137 state, when logoff message received sm goes here.
    LOGOFF = "LOGOFF"
    LOGOFF2 = "LOGOFF2"

    STATES = [State(NO_STATE, 'reset_state'),
              State(DISABLED, 'disabled_state'),
              State(INITIALIZE, 'initialize_state'),
              State(IDLE, 'idle_state'),
              State(RECEIVED, 'received_state'),
              State(INTEGRITY_CHECK, 'integrity_check_state'),
              State(METHOD_RESPONSE, 'method_response_state'),
              State(METHOD_REQUEST, 'method_request_state'),
              State(PROPOSE_METHOD, 'propose_method_state'),
              State(SELECT_ACTION, 'select_action_state'),
              State(SEND_REQUEST, 'send_request_state'),
              State(DISCARD, 'discard_state'),
              State(NAK, 'nak_state'),
              State(RETRANSMIT, 'retransmit_state'),
              State(SUCCESS, 'success_state'),
              State(FAILURE, 'failure_state'),
              State(TIMEOUT_FAILURE, 'timeout_failure_state'),
              State(INITIALIZE_PASSTRHOUGH, 'initialize_passthrough_state'),
              State(IDLE2, 'idle2_state'),
              State(RECEIVED2, 'received2_state'),
              State(AAA_IDLE, 'aaa_idle_state'),
              State(AAA_REQUEST, 'aaa_request_state'),
              State(AAA_RESPONSE, 'aaa_response_state'),
              State(SEND_REQUEST2, 'send_request2_state'),
              State(DISCARD2, 'discard2_state'),
              State(RETRANSMIT2, 'retransmit2_state'),
              State(SUCCESS2, 'success2_state'),
              State(FAILURE2, 'failure2_state'),
              State(TIMEOUT_FAILURE2, 'timeout_failure2_state'),
              State(LOGOFF, 'logoff_state'),
              State(LOGOFF2, 'logoff2_state')
              ]

    TRANSITIONS = [{'trigger': 'process', 'source': '*', 'dest': DISABLED,
                    'unless': ['is_port_enabled']},
                   {'trigger': 'process', 'source': '*', 'dest': INITIALIZE,
                    'conditions': ['is_port_enabled',
                                   'is_eap_restart']},
                   {'trigger': 'process', 'source': DISABLED, 'dest': NO_STATE,
                    'conditions': ['is_port_enabled']},
                   {'trigger': 'process', 'source': INITIALIZE, 'dest': SELECT_ACTION},
                   {'trigger': 'process', 'source': SELECT_ACTION, 'dest': PROPOSE_METHOD,
                    'unless': ['is_decision_failure',
                               'is_decision_passthrough',
                               'is_decision_success']},
                   {'trigger': 'process', 'source': SELECT_ACTION, 'dest': FAILURE,
                    'conditions': ['is_decision_failure']},
                   {'trigger': 'process', 'source': SELECT_ACTION, 'dest': SUCCESS,
                    'conditions': ['is_decision_success']},
                   {'trigger': 'process', 'source': SELECT_ACTION, 'dest': INITIALIZE_PASSTRHOUGH,
                    'conditions': ['is_decision_passthrough']},
                   {'trigger': 'process', 'source': PROPOSE_METHOD, 'dest': METHOD_REQUEST},
                   {'trigger': 'process', 'source': METHOD_REQUEST, 'dest': SEND_REQUEST},
                   {'trigger': 'process', 'source': SEND_REQUEST, 'dest': IDLE},
                   {'trigger': 'process', 'source': IDLE, 'dest': RETRANSMIT,
                    'conditions': ['is_retrans_while_equal_0']},
                   {'trigger': 'process', 'source': IDLE, 'dest': RECEIVED,
                    'conditions': ['is_eap_resp']},
                   {'trigger': 'process', 'source': RETRANSMIT, 'dest': TIMEOUT_FAILURE,
                    'conditions': ['is_retrans_count_greater_max_retrans']},
                   {'trigger': 'process', 'source': RETRANSMIT, 'dest': IDLE,
                    'unless': ['is_retrans_count_greater_max_retrans']},
                   {'trigger': 'process', 'source': RECEIVED, 'dest': NAK,
                    'conditions': ['is_enter_nak']},
                   {'trigger': 'process', 'source': RECEIVED, 'dest': INTEGRITY_CHECK,
                    'conditions': ['is_enter_integrity_check']},
                   {'trigger': 'process', 'source': RECEIVED, 'dest': DISCARD,
                    'unless': ['is_enter_nak', 'is_enter_integrity_check']},
                   {'trigger': 'process', 'source': DISCARD, 'dest': IDLE},
                   {'trigger': 'process', 'source': NAK, 'dest': SELECT_ACTION},
                   {'trigger': 'process', 'source': INTEGRITY_CHECK, 'dest': DISCARD,
                    'conditions': ['is_ignore']},
                   {'trigger': 'process', 'source': INTEGRITY_CHECK, 'dest': METHOD_RESPONSE,
                    'unless': ['is_ignore']},
                   {'trigger': 'process', 'source': METHOD_RESPONSE, 'dest': SELECT_ACTION,
                    'conditions': ['is_method_state_equal_end']},
                   {'trigger': 'process', 'source': METHOD_RESPONSE, 'dest': METHOD_REQUEST,
                    'unless': ['is_method_state_equal_end']},

                   {'trigger': 'process', 'source': INITIALIZE_PASSTRHOUGH, 'dest': AAA_IDLE,
                    'conditions': ['is_current_id_none']},
                   {'trigger': 'process', 'source': INITIALIZE_PASSTRHOUGH, 'dest': AAA_REQUEST,
                    'unless': ['is_current_id_none']},
                   {'trigger': 'process', 'source': AAA_REQUEST, 'dest': AAA_IDLE},
                   {'trigger': 'process', 'source': AAA_IDLE, 'dest': TIMEOUT_FAILURE2,
                    'conditions': ['is_aaa_timeout']},
                   {'trigger': 'process', 'source': AAA_IDLE, 'dest': FAILURE2,
                    'conditions': ['is_aaa_fail']},
                   {'trigger': 'process', 'source': AAA_IDLE, 'dest': SUCCESS2,
                    'conditions': ['is_aaa_success']},
                   {'trigger': 'process', 'source': AAA_IDLE, 'dest': AAA_RESPONSE,
                    'conditions': ['is_aaa_eap_req']},
                   {'trigger': 'process', 'source': AAA_IDLE, 'dest': DISCARD2,
                    'conditions': ['is_aaa_eap_no_req']},
                   {'trigger': 'process', 'source': DISCARD2, 'dest': IDLE2},
                   {'trigger': 'process', 'source': AAA_RESPONSE, 'dest': SEND_REQUEST2},
                   {'trigger': 'process', 'source': SEND_REQUEST2, 'dest': IDLE2},
                   {'trigger': 'process', 'source': IDLE2, 'dest': RETRANSMIT2,
                    'conditions': ['is_retrans_while_equal_0']},
                   {'trigger': 'process', 'source': IDLE2, 'dest': RECEIVED2,
                    'conditions': ['is_eap_resp']},
                   {'trigger': 'process', 'source': RETRANSMIT2, 'dest': TIMEOUT_FAILURE2,
                    'conditions': ['is_retrans_count_greater_max_retrans']},
                   {'trigger': 'process', 'source': RETRANSMIT2, 'dest': IDLE2,
                    'unless': ['is_retrans_count_greater_max_retrans']},
                   {'trigger': 'process', 'source': RECEIVED2, 'dest': AAA_REQUEST,
                    'conditions': ['is_rx_resp', 'is_resp_id_equal_current_id']},
                   {'trigger': 'process', 'source': RECEIVED2, 'dest': DISCARD2,
                    'conditions': ['is_enter_discard2']},

                   {'trigger': 'process', 'source': SUCCESS, 'dest': LOGOFF,
                    'conditions': ['is_logoff']},
                   {'trigger': 'process', 'source': SUCCESS2, 'dest': LOGOFF2,
                    'conditions': ['is_logoff']},
                   ]

    # RFC 4137
    MAX_RETRANS = 5  # Configurable  max for retransmissions before aborting.

    # Variables (AAA Interface to Full Authenticator)
    aaa_eap_req = None        # bool
    aaa_eap_no_req = None      # bool
    aaa_success = None       # bool
    aaa_fail = None          # bool
    aaa_eap_req_data = None    # EAP packet
    aaa_eap_key_data = None    # EAP Key
    aaa_eap_key_available = None   # bool
    aaa_method_timeout = None     # integer or NONE

    # Variables (Full Authenticator to AAA Interface)
    aaa_eap_resp = None       # bool
    aaa_eap_resp_data = None   # EAP Packet
    aaa_identity = None      # EAP Packet
    aaa_timeout = None      # bool

    # Stand-Alone Authenticator State Machine Local Variables
    current_method = None    # EAP type
    current_id = None        # integer
    method_state = None      # enum
    retrans_count = None     # integer
    last_req_data = None      # EAP packet
    method_timeout = None    # integer
    logoff = None           # bool
    # Non RFC 4137
    override_current_id = None

    # Lower Later  to Stand-Alone Authenticator
    eap_resp = None      # bool
    eap_resp_data = None  # EAP Packet
    port_enabled = None  # bool
    retrans_while = None     # integer
    eap_restart = None   # bool
    eap_srtt = None      # integer
    eap_rttvar = None    # integer

    # Stand-Alone authenticator to Lower Layer
    eap_req = None       # bool
    eap_no_req = None     # bool
    eap_success = None   # bool
    eap_fail = None      # bool
    eap_timeout = None   # bool
    eap_req_data = None   # EAP Packet
    eap_key_data = None   # EAP Key
    eap_key_available = None  # bool
    # Non RFC 4137
    eap_logoff = None    # bool

    # short term local variables (not maintained between packets)
    rx_resp = None
    resp_id = None
    resp_method = None
    ignore = None
    decision = None

    def __init__(self, eap_output_queue, radius_output_queue, src_mac, timer_scheduler,
                 auth_handler, failure_handler, logoff_handler, log_prefix):
        """

        Args:
            auth_handler (callable): callable that takes input of src_mac. Called on EAP-Success.
            eap_output_queue (Queue): where to put Messages to send to supplicant
            failure_handler (callable): callable that takes input of src_mac. Called on EAP-Failure.
            logoff_handler (callable): callable that takes input of src_mac. Called on EAP-Logoff.
            radius_output_queue (Queue): where to put Messages to send to AAA server
            src_mac (MacAddress): MAC address this statemachine (sm) belongs to.
            timer_scheduler (Scheduler): where to put timer events. (useful for Retransmits)
        """
        self.eap_output_messages = eap_output_queue
        self.radius_output_messages = radius_output_queue
        self.src_mac = src_mac
        self.timer_scheduler = timer_scheduler
        self.auth_handler = auth_handler
        self.failure_handler = failure_handler
        self.logoff_handler = logoff_handler

        self.machine = Machine(model=self, states=FullEAPStateMachine.STATES,
                               transitions=FullEAPStateMachine.TRANSITIONS,
                               queued=True,
                               initial=FullEAPStateMachine.NO_STATE)

        # TODO dynamically assign this or make a way to give it multiple methods
        # and self.m is the one currently in use.
        # if we want to deal with each method locally.
        self.m = MPassthrough()  # pylint: disable=invalid-name
        self.logger = get_logger(log_prefix)

        self.eap_restart = True
        self.port_enabled = True

    def is_eap_restart(self):
        return self.eap_restart

    def is_port_enabled(self):
        return self.port_enabled

    def is_decision_failure(self):
        return self.decision == Decision.FAILURE

    def is_decision_success(self):
        return self.decision == Decision.SUCCESS

    def is_decision_passthrough(self):
        return self.decision == Decision.PASSTHROUGH

    def is_retrans_while_equal_0(self):
        return self.retrans_while == 0

    def is_eap_resp(self):
        return self.eap_resp

    def is_retrans_count_greater_max_retrans(self):
        return self.retrans_count > self.MAX_RETRANS

    def is_enter_nak(self):
        return self.rx_resp and self.resp_id == self.current_id \
               and (self.resp_method in (MethodState.NAK, MethodState.EXPANDED_NAK)) \
               and self.method_state == MethodState.PROPOSED

    def is_enter_integrity_check(self):
        return self.rx_resp and self.resp_id == self.current_id \
               and self.resp_method == self.current_method

    def is_ignore(self):
        return self.ignore

    def is_method_state_equal_end(self):
        return self.method_state == MethodState.END

    def is_current_id_none(self):
        return self.current_id is None

    def is_aaa_timeout(self):
        return self.aaa_timeout

    def is_aaa_fail(self):
        return self.aaa_fail

    def is_aaa_success(self):
        return self.aaa_success

    def is_aaa_eap_req(self):
        return self.aaa_eap_req

    def is_aaa_eap_no_req(self):
        return self.aaa_eap_no_req

    def is_rx_resp(self):
        return self.rx_resp is not None

    def is_resp_id_equal_current_id(self):
        return self.resp_id == self.current_id

    def is_enter_discard2(self):
        return self.rx_resp is None or self.resp_id != self.current_id

    def is_logoff(self):
        return self.logoff

    def get_id(self):
        """Determines the identifier value chosen by the AAA server for the current EAP request.
         The return value is an integer."""
        return self.eap_req_data.message_id

    def calculate_timeout(self, retrans_count, eap_srtt, eap_rttvar, method_timeout):  # pylint: disable=unused-argument
        """https://tools.ietf.org/html/rfc3748#section-4.3
        Args:
            retrans_count:
            eap_srtt:
            eap_rttvar:
            method_timeout:

        Returns:
            Milliseconds"""
        # TODO actually implement.
        return self.DEFAULT_TIMEOUT

    def parse_eap_resp(self):
        """
        Returns:
            int, int, EAP Type (str)
        """
        eap = self.eap_resp_data
        resp_method = None

        _id = eap.message_id

        if isinstance(eap, IdentityMessage):
            resp_method = MethodState.IDENTITY
        # RFC 4137 #section 5.4 says eap.code should actually be a bool
        eap_code = getattr(eap, 'code', None)
        return eap_code, _id, resp_method

    def build_success(self):
        """Creates an EAP Sucecss Pakcet. Returns an EAP packet"""
        return SuccessMessage(self.src_mac, self.current_id)

    def build_failure(self):
        """Creates an EAP Failure Packet. Returns an EAP packet"""
        return FailureMessage(self.src_mac, self.current_id)

    def next_id(self):
        """Determines the next identifier value to use, based on the previous one.
        Returns:
            integer"""
        if self.current_id is None:
            # I'm assuming we cant have ids wrap around in the same series.
            #  so the 200 provides a large buffer.
            return random.randint(0, 200)
        _id = self.current_id + 1
        # not tested
        if _id > 255:
            return random.randint(0, 200)
        return _id

    @log_method
    def disabled_state(self):
        """The authenticator is disabled until the port is enabled by the lower layer"""
        # DISABLED does not do anything.
        pass

    @log_method
    def propose_method_state(self):
        self.current_method = Policy.get_next_method(self.eap_resp_data)
        self.m.init(self.src_mac)
        if self.current_method == "IDENTITY" or self.current_method == "NOTIFICATION":
            self.method_state = MethodState.CONTINUE
        else:
            self.method_state = MethodState.PROPOSED

    @log_method
    def failure_state(self):
        self.eap_req_data = self.build_failure()
        self.eap_fail = True

    @log_method
    def timeout_failure_state(self):
        self.eap_timeout = True

    @log_method
    def success_state(self):
        self.eap_req_data = self.build_success()
        if self.eap_key_data:
            self.eap_key_available = True
        self.eap_success = True

    @log_method
    def initialize_state(self):
        """Initializes variables when the state machine is activated"""
        self.current_id = None
        if self.override_current_id:
            self.current_id = self.override_current_id
        self.override_current_id = None
        self.eap_success = False
        self.eap_fail = False
        self.eap_timeout = False
        self.eap_key_data = None
        self.eap_restart = False

        self.eap_logoff = False

        self.radius_state_attribute = None

    @log_method
    def idle_state(self):
        """The state machine spends most of its time here, waiting for something to happen"""
        self.retrans_while = self.calculate_timeout(self.retrans_count, self.eap_srtt,
                                                    self.eap_rttvar, self.method_timeout)

    @log_method
    def received_state(self):
        """This state is entered when an EAP packet is received. The packet header is parsed here"""
        self.rx_resp, self.resp_id, self.resp_method = self.parse_eap_resp()

    @log_method
    def select_action_state(self):
        self.decision = Policy.get_decision(self.eap_resp_data)

    @log_method
    def method_response_state(self):
        self.m.process(self.eap_resp_data)
        if self.m.is_done():
            Policy.update()
            self.eap_key_data = self.m.get_key()
            self.method_state = MethodState.END
        else:
            self.method_state = MethodState.CONTINUE

    @log_method
    def discard_state(self):
        self.eap_resp = False
        self.eap_no_req = True

    @log_method
    def integrity_check_state(self):
        self.ignore = self.m.check(self.eap_resp_data)

    @log_method
    def nak_state(self):
        self.m.reset()
        Policy.update()

    @log_method
    def retransmit_state(self):
        self.retrans_count += 1
        if self.retrans_count <= self.MAX_RETRANS:
            self.eap_req_data = self.last_req_data
            self.eap_req = True

    @log_method
    def send_request_state(self):
        self.retrans_count = 0
        self.last_req_data = self.eap_req_data
        self.eap_resp = False
        self.eap_req = True

    @log_method
    def method_request_state(self):
        self.current_id = self.next_id()
        self.eap_req_data = self.m.build_req(self.current_id)
        self.method_timeout = self.m.get_timeout()

    @log_method
    def initialize_passthrough_state(self):
        self.aaa_eap_resp = None
        # non rfc behaviour. if
        self.rx_resp, self.resp_id, self.resp_method = self.parse_eap_resp()

    @log_method
    def idle2_state(self):
        self.retrans_while = self.calculate_timeout(self.retrans_count, self.eap_srtt,
                                                    self.eap_rttvar, self.method_timeout)

    @log_method
    def received2_state(self):
        self.rx_resp, self.resp_id, self.resp_method = self.parse_eap_resp()


    @log_method
    def aaa_request_state(self):
        if self.resp_method == MethodState.IDENTITY:
            self.aaa_identity = self.eap_resp_data
        self.aaa_eap_resp_data = self.eap_resp_data

    @log_method
    def aaa_idle_state(self):
        self.aaa_fail = False
        self.aaa_success = False
        self.aaa_eap_req = False
        self.aaa_eap_no_req = False
        self.aaa_eap_resp = True

    @log_method
    def aaa_response_state(self):
        self.eap_req_data = self.aaa_eap_req_data
        self.current_id = self.get_id()
        self.method_timeout = self.aaa_method_timeout

    @log_method
    def send_request2_state(self):
        self.retrans_count = 0
        self.last_req_data = self.eap_req_data
        self.eap_resp = False
        self.eap_req = True

    @log_method
    def discard2_state(self):
        self.eap_resp = False
        self.eap_no_req = True

    @log_method
    def retransmit2_state(self):
        self.retrans_count += 1
        if self.retrans_count <= self.MAX_RETRANS:
            self.eap_req_data = self.last_req_data
            self.eap_req = True

    @log_method
    def success2_state(self):
        self.eap_req = True
        self.eap_req_data = self.aaa_eap_req_data
        self.eap_key_data = self.aaa_eap_key_data
        self.eap_key_available = self.aaa_eap_key_available
        self.eap_success = True

    @log_method
    def failure2_state(self):
        self.eap_req = True
        self.eap_req_data = self.aaa_eap_req_data
        self.eap_fail = True

    @log_method
    def timeout_failure2_state(self):
        self.eap_timeout = True

    @log_method
    def logoff_state(self):
        self.eap_success = False
        self.eap_logoff = True

    @log_method
    def logoff2_state(self):
        self.eap_success = False
        self.eap_logoff = True

    @log_method
    def reset_state(self):
        self.initialize_state()
        self.aaa_eap_req_data = None
        self.aaa_eap_key_data = None
        self.eap_req_data = None
        self.eap_key_data = None
        self.eap_success = False
        self.aaa_success = False

    def handle_message_received(self):
        """Main state machine loop"""

        self.rx_resp = None
        self.resp_id = None
        self.resp_method = None
        self.ignore = None
        self.decision = None

        last_state = None
        while self.state != last_state:
            last_state = self.state
            self.process()  # pylint: disable=no-member # pytype: disable=attribute-error

    def lower_layer_reset(self):
        """Sets variables that are meant to be set by the lower layer
        RFC4137 5.1.2 (standalone authenticator to Lower Layer)"""
        self.eap_req = False
        self.eap_no_req = False
        self.eap_success = False
        self.eap_fail = False
        self.eap_timeout = False

        self.logoff = False

        self.aaa_eap_resp = False
        self.aaa_timeout = False

    def strip_eap_from_radius_packet(self, radius):
        """Build a EventRadiusMessageReceived from a radius message"""
        eap_msg_attribute = radius.attributes.find(radius_attributes.EAPMessage.DESCRIPTION)
        eap_msg = eap_msg_attribute.data_type.data()
        state = radius.attributes.find(radius_attributes.State.DESCRIPTION)
        self.logger.info("radius EAP: %s", eap_msg)
        return EventRadiusMessageReceived(eap_msg, state, radius.attributes.to_dict())

    def event(self, event):
        """Processes an event.
        Output is via the eap/radius queue. and again will be of type ***Message.
        Args:
            event: should have message attribute which is of the ***Message types
            (e.g. SuccessMessage, IdentityMessage,...)
        """

        # TODO remove and refactor code - Just placing here to separate main pipeline for internals of SM
        if (isinstance(event, EventPreemptiveEAPResponseMessageReceived)
                and event.preemptive_eap_id != self.current_id):
            self.logget.info("Resetting eap due to recieved response to preemtive request")
            self.eap_restart = True
            self.override_current_id = event.preemptive_eap_id

        if isinstance(event, EventRadiusMessageReceived) and isinstance(event.message, RadiusPacket):
            event = self.strip_eap_from_radius_packet(event.message)

        self.lower_layer_reset()
        self.logger.info("full state machine received event: %s", event)
        # 'Lower Layer' shim
        if isinstance(event, EventMessageReceived):
            self.message_event_received(event)

        elif isinstance(event, EventTimerExpired):
            if self.timer_expired_event_received(event):
                return

        elif isinstance(event, EventPortStatusChange):
            self.port_status_event_received(event)
        elif isinstance(event, EventSessionTimeout):
            self.session_timeout_event_received()

        self.handle_message_received()
        self.logger.info('end state: %s', self.state)

        if self.eap_req:
            if (hasattr(self.eap_req_data, 'code') and self.eap_req_data.code == Eap.REQUEST) \
                    or isinstance(self.eap_req_data, (SuccessMessage, FailureMessage)):
                self.logger.info("outputting eap, '%s', src: '%s' port_id: '%s'",
                                 self.eap_req_data, self.src_mac, self.port_id_mac)
                self.eap_output_messages.put_nowait(
                    EapQueueMessage(self.eap_req_data, self.src_mac, self.port_id_mac))
                self.sent_count += 1
                self.set_timer(self.retrans_while)
            # not tested
            else:
                self.logger.error('cant find code --- %s', self.eap_req_data)
            self.eap_req = False

        if self.aaa_eap_resp and self.aaa_eap_resp_data:
            if self.aaa_eap_resp_data.code == Eap.RESPONSE:
                self.logger.info('outputing radius')
                self.radius_output_messages.put_nowait(
                    RadiusQueueMessage(self.aaa_eap_resp_data, self.src_mac,
                                       self.aaa_identity.identity,
                                       self.radius_state_attribute, self.port_id_mac))

                self.sent_count += 1
                self.set_timer(self.RADIUS_RETRANSMIT_TIMEOUT)
            self.aaa_eap_resp = False
        # not tested
        elif self.aaa_eap_resp:
            self.logger.error("aaa_eap_resp is true. but data is false. This should never happen")

        if self.eap_success:
            self.handle_success()

        if self.eap_fail:
            self.logger.info('oh authentication not successful %s', self.src_mac)
            self.failure_handler(self.src_mac, str(self.port_id_mac))

        if self.eap_logoff:
            self.handle_logoff()

    def handle_logoff(self):
        """Notify the logoff callback"""
        self.logger.info('client is logging off %s', self.src_mac)
        self.logoff_handler(self.src_mac, str(self.port_id_mac))
        if self.session_timeout_job:
            self.session_timeout_job.cancel()

    def handle_success(self):
        """Notify the success callback and sets a timer event to expire this session"""
        self.logger.info('Yay authentication successful %s %s',
                         self.src_mac, self.aaa_identity.identity)
        self.auth_handler(self.src_mac, str(self.port_id_mac),
                          self.session_timeout, self.radius_tunnel_private_group_id, self.filter_id)
        self.aaa_eap_resp_data = None

        # new authentication so cancel the old session timeout event
        if self.session_timeout_job:
            self.session_timeout_job.cancel()

        self.session_timeout_job = self.timer_scheduler.call_later(self.session_timeout,
                                                                   self.event,
                                                                   EventSessionTimeout(self))

    def session_timeout_event_received(self):
        """process session timeout event"""
        self.logoff = True

    def port_status_event_received(self, event):
        """Sets variables for the port status change (link up/down) being received.
        Args:
            event (EventPortStatusChange):
        """
        self.port_enabled = event.port_status

    def timer_expired_event_received(self, event):
        """Check if the event has been replied to. and set variables.
        Args:
            event (EventTimerExpired): event to process

        Returns:
            True if this event is being ignored and no further processing is required.
            Otherwise False.
        """
        # TODO Should this still log all ExpiredTimerEvents when none are cancelled?
        self.logger.info("Expired Timer Event Received")
        if self.sent_count == event.sent_count:
            self.logger.debug("processing timer event. haven't received a reply. %s %s",
                              self.sent_count, event.sent_count)

            if self.state == self.AAA_IDLE:
                self.aaa_timeout = True
            if self.state == self.IDLE2 or self.state == self.IDLE:
                self.retrans_while = 0

            return False
        self.logger.debug("ignoring timer event, already received a reply.")
        return True

    def message_event_received(self, event):
        """Sets variables for the Eap message being received.
        Args:
            event (EventMessageReceived): event being processed.
        """
        message = event.message
        self.logger.info('type: %s, message %s ', type(message), message)
        if event.port_id:
            self.port_id_mac = event.port_id

        if isinstance(message, EapolStartMessage) or \
                (self.state in (FullEAPStateMachine.TIMEOUT_FAILURE,
                                FullEAPStateMachine.TIMEOUT_FAILURE2) and
                 isinstance(message, EapMessage) and message.code == Eap.RESPONSE
                ):
            self.eap_restart = True
        elif isinstance(message, EapolLogoffMessage):
            self.logoff = True

        if isinstance(event, EventRadiusMessageReceived):
            self.process_radius_message(event)
        else:
            self.eap_resp_data = message
            self.eap_resp = True

    def process_radius_message(self, event):
        """Process radius message (set and extract radius specific variables)"""
        self.eap_resp_data = None
        self.eap_resp = False
        self.logger.debug('radius attributes %s', event.attributes)
        self.radius_state_attribute = event.state
        self.aaa_eap_req = True
        self.aaa_eap_req_data = event.message
        self.logger.debug('sm ev.msg: %s', self.aaa_eap_req_data)
        if isinstance(self.aaa_eap_req_data, SuccessMessage):
            self.logger.debug("aaaSuccess")
            self.aaa_success = True
        if isinstance(self.aaa_eap_req_data, FailureMessage):
            self.logger.debug("aaaFail")
            self.aaa_fail = True
        self.logger.debug('radius event %s', event.__dict__)
        self.set_vars_from_radius(event.attributes)

    def set_vars_from_radius(self, attributes):
        """
        Set the statemachine variables from attributes received in the radius message.
        If variable does not exist in the radius message, it is reset to the default
        Args:
            attributes (dict):  attributes to be set.
        """
        self.session_timeout = self.DEFAULT_SESSION_TIMEOUT
        self.radius_tunnel_private_group_id = None
        self.filter_id = None

        if attributes:
            self.session_timeout = attributes.get(radius_attributes.SessionTimeout.DESCRIPTION,
                                                  self.DEFAULT_SESSION_TIMEOUT)
            self.radius_tunnel_private_group_id = attributes.get(radius_attributes.TunnelPrivateGroupID.DESCRIPTION,
                                                                 None)
            self.filter_id = attributes.get(radius_attributes.FilterId.DESCRIPTION,
                                                                 None)
            if self.radius_tunnel_private_group_id:
                self.radius_tunnel_private_group_id = self.radius_tunnel_private_group_id.decode('utf-8')
        # TODO could also set filter-id/vlans/acls here.

    def set_timer(self, timeout):
        """Sets a timer to trigger a retransmit if no packet received.
        """
        # These messages should not expect a reply, so set the timer.
        if self.state not in [self.SUCCESS, self.SUCCESS2,
                              self.FAILURE, self.FAILURE2,
                              self.TIMEOUT_FAILURE, self.TIMEOUT_FAILURE2]:
            self.timer_scheduler.call_later(timeout,
                                            self.event,
                                            EventTimerExpired(self, self.sent_count))
            # TODO could cancel the scheduled events when
            # they're no longer needed (i.e. response received)

    def is_in_progress(self):
        return self.state not in [FullEAPStateMachine.LOGOFF, FullEAPStateMachine.LOGOFF2,
                                  FullEAPStateMachine.DISABLED, FullEAPStateMachine.NO_STATE,
                                  FullEAPStateMachine.FAILURE, FullEAPStateMachine.FAILURE2,
                                  FullEAPStateMachine.TIMEOUT_FAILURE,
                                  FullEAPStateMachine.TIMEOUT_FAILURE2,]
                                  # FullEAPStateMachine.SUCCESS, FullEAPStateMachine.SUCCESS2]

    def is_success(self):
        return self.state in [FullEAPStateMachine.SUCCESS, FullEAPStateMachine.SUCCESS2]



"""Handle activity that is not EAP on the same EAP interface"""

import struct
from fcntl import ioctl
from eventlet.green import socket

from chewie.mac_address import MacAddress

class ActivitySocket:
    """Handle the RADIUS socket"""
    SIOCGIFINDEX = 0x8933
    PACKET_MR_PROMISC = 1
    IP_ETHERTYPE = 0x0800
    SOL_PACKET = 263
    PACKET_ADD_MEMBERSHIP = 1

    DHCP_UDP_SRC = 68
    DHCP_UDP_DST = 67
    UDP_IPTYPE = b'\x11'
    EAP_ADDRESS = MacAddress.from_string("01:80:c2:00:00:03")

    def __init__(self, interface_name):
        self.socket = None
        self.interface_name = interface_name
        self.interface_index = None

    def setup(self):
        """Set up the socket"""
        self.open()
        self.get_interface_index()
        self.set_interface_promiscuous()

    def send(self, data):
        """Not Implemented -- This socket is purely for Listening"""
        raise NotImplementedError('Attempted to send data down the activity socket')

    def receive(self):
        """Receive activity from supplicant-facing socket"""
        # Skip all packets that are not DHCP requests
        while True:
            ret_val = self.socket.recv(4096)

            if ret_val[23:24] == self.UDP_IPTYPE:
                src_port = struct.unpack('>H', ret_val[34:36])[0]
                dst_port = struct.unpack('>H', ret_val[36:38])[0]

                if src_port == self.DHCP_UDP_SRC and dst_port == self.DHCP_UDP_DST:
                    return ret_val

    def open(self):
        """Listen on the Socket for any form of Eth() / IP() frames """
        self.socket = socket.socket(socket.PF_PACKET, socket.SOCK_RAW,  # pylint: disable=no-member
                                    socket.htons(self.IP_ETHERTYPE))    # pylint: disable=no-member
        self.socket.bind((self.interface_name, 0))

    def get_interface_index(self):
        """Get the interface index of the Socket"""
        # http://man7.org/linux/man-pages/man7/netdevice.7.html
        request = struct.pack('16sI', self.interface_name.encode("utf-8"), 0)
        response = ioctl(self.socket, self.SIOCGIFINDEX, request)
        _ifname, self.interface_index = struct.unpack('16sI', response)

    def set_interface_promiscuous(self):
        """Sets the activity interface to be able to receive messages with port_id in mac_dst"""
        request = struct.pack("IHH8s", self.interface_index, self.PACKET_MR_PROMISC,
                              len(self.EAP_ADDRESS.address), self.EAP_ADDRESS.address)

        self.socket.setsockopt(self.SOL_PACKET, self.PACKET_ADD_MEMBERSHIP, request)



"""Handle the EAP socket"""

import struct
from fcntl import ioctl
from eventlet.green import socket

from chewie.mac_address import MacAddress


class EapSocket:
    """Handle the EAP socket"""
    SIOCGIFINDEX = 0x8933
    PACKET_MR_PROMISC = 1
    SOL_PACKET = 263
    PACKET_ADD_MEMBERSHIP = 1
    EAP_ADDRESS = MacAddress.from_string("01:80:c2:00:00:03")

    def __init__(self, interface_name):
        self.socket = None
        self.interface_index = None
        self.interface_name = interface_name

    def setup(self):
        """Set up the socket"""
        self.open()
        self.get_interface_index()
        self.set_interface_promiscuous()

    def send(self, data):
        """send on eap socket.
            data (bytes): data to send"""
        self.socket.send(data)

    def receive(self):
        """receive from eap socket"""
        return self.socket.recv(4096)

    def open(self):
        """Setup EAP socket"""

        self.socket = socket.socket(socket.PF_PACKET, socket.SOCK_RAW,  # pylint: disable=no-member
                                    socket.htons(0x888e))               # pylint: disable=no-member
        self.socket.bind((self.interface_name, 0))

    def get_interface_index(self):
        """Get the interface index of the EAP Socket"""
        # http://man7.org/linux/man-pages/man7/netdevice.7.html
        request = struct.pack('16sI', self.interface_name.encode("utf-8"), 0)
        response = ioctl(self.socket, self.SIOCGIFINDEX, request)
        _ifname, self.interface_index = struct.unpack('16sI', response)

    def set_interface_promiscuous(self):
        """Sets the EAP interface to be able to receive EAP messages"""
        request = struct.pack("IHH8s", self.interface_index, self.PACKET_MR_PROMISC,
                              len(self.EAP_ADDRESS.address), self.EAP_ADDRESS.address)
        self.socket.setsockopt(self.SOL_PACKET, self.PACKET_ADD_MEMBERSHIP, request)



import logging
import tempfile
import unittest
from queue import Queue

from chewie.ethernet_packet import EthernetPacket
from chewie.event import EventMessageReceived, EventRadiusMessageReceived
from chewie.mac_address import MacAddress
from chewie.radius import RadiusAccessReject, RadiusAccessAccept
from chewie.state_machines.mab_state_machine import MacAuthenticationBypassStateMachine


# TODO Remove and create a 'test state machine' class
def check_counters(_func=None, *,
                   expected_auth_counter=0, expected_failure_counter=0):
    """Decorator to check the handlers have been called the
     correct number of times at the end of each test"""

    def decorator_check_counters(func):
        def wrapper(self):
            start_auth_counter = self.auth_counter
            start_failure_counter = self.failure_counter
            ret = func(self)
            self.assertEqual(self.auth_counter,
                             start_auth_counter + expected_auth_counter)
            self.assertEqual(self.failure_counter,
                             start_failure_counter + expected_failure_counter)
            return ret

        return wrapper

    if _func is None:
        return decorator_check_counters
    else:
        return decorator_check_counters(_func)


class MABStateMachineTest(unittest.TestCase):
    PORT_ID_MAC = MacAddress.from_string("00:00:00:00:00:01")

    def setUp(self):
        # Build the state machine
        logger = logging.getLogger()
        logger.level = logging.DEBUG
        self.log_file = tempfile.NamedTemporaryFile()
        logger.addHandler(logging.FileHandler(self.log_file.name))

        self.radius_output_queue = Queue()
        self.timer_scheduler = None
        self.src_mac = MacAddress.from_string("00:12:34:56:78:90")
        log_prefix = "chewie.SM - port: %s, client: %s" % (self.src_mac, self.PORT_ID_MAC)

        self.sm = MacAuthenticationBypassStateMachine(self.radius_output_queue,
                                                      self.src_mac, self.timer_scheduler,
                                                      self.auth_handler, self.failure_handler,
                                                      log_prefix)

        self.auth_counter = 0
        self.failure_counter = 0

    # pylint: disable=unused-argument
    def auth_handler(self, client_mac, port_id_mac, timer, *arg, **kwargs):
        self.auth_counter += 1
        print('Successful auth from MAC %s' % str(client_mac))

    def failure_handler(self, client_mac, port_id_mac):  # pylint: disable=unused-argument
        self.failure_counter += 1
        print('failure from MAC %s' % str(client_mac))

    def receive_eth_packet(self):
        """Receive Ethernet Frame and send to State Machine"""
        rad_queue_size = self.radius_output_queue.qsize()
        message = EthernetPacket(self.PORT_ID_MAC, str(self.src_mac), 0x888e, "")
        self.sm.event(EventMessageReceived(message, self.PORT_ID_MAC))
        self.assertEqual(self.radius_output_queue.qsize(), rad_queue_size + 1)

    def receive_radius_accept(self):
        """Receive Radius AccessAccept Packet and send to State Machine"""
        rad_queue_size = self.radius_output_queue.qsize()
        message = RadiusAccessAccept(1, 1, [])
        self.sm.event(EventRadiusMessageReceived(message, None, []))
        self.assertEqual(self.radius_output_queue.qsize(), rad_queue_size)

    def receive_radius_reject(self):
        """Receive Radius AccessReject Packet and send to State Machine"""
        rad_queue_size = self.radius_output_queue.qsize()
        message = RadiusAccessReject(1, 1, [])
        self.sm.event(EventRadiusMessageReceived(message, None, []))
        self.assertEqual(self.radius_output_queue.qsize(), rad_queue_size)

    def test_smoke_test_send_request(self):
        """Smoke Test Send RADIUS MAB Request on Receiving DHCP Activity"""
        self.sm.port_enabled = True

        self.receive_eth_packet()
        radius_output = self.radius_output_queue.get_nowait()
        self.assertEqual(self.sm.AAA_IDLE, self.sm.state)
        self.assertEqual(radius_output[0], str(self.src_mac))

    @check_counters(expected_auth_counter=1)
    def test_smoke_test_success_radius(self):
        """Smoke Test successful RADIUS Request"""
        self.sm.port_enabled = True
        self.receive_eth_packet()

        radius_output = self.radius_output_queue.get_nowait()
        self.assertEqual(self.sm.AAA_IDLE, self.sm.state)
        self.assertEqual(radius_output[0], str(self.src_mac))

        self.receive_radius_accept()
        self.assertEqual(self.sm.AAA_SUCCESS, self.sm.state)

    @check_counters(expected_failure_counter=1)
    def test_smoke_test_fail_radius(self):
        """Smoke Test incorrect details sent to RADIUS Server"""
        self.sm.port_enabled = True
        self.receive_eth_packet()

        radius_output = self.radius_output_queue.get_nowait()
        self.assertEqual(self.sm.AAA_IDLE, self.sm.state)
        self.assertEqual(radius_output[0], str(self.src_mac))

        self.receive_radius_reject()
        self.assertEqual(self.sm.AAA_FAILURE, self.sm.state)



"""Various Events used by Chewie"""


# pylint: disable=too-few-public-methods
class Event:
    """Base event class"""
    TIMER_EXPIRED = 1
    MESSAGE_RECEIVED = 2
    SHUTDOWN = 3
    PORT_ENABLED = 4
    PORT_DISABLED = 5


class EventTimerExpired(Event):
    """Used when a timer has expired."""

    def __init__(self, state_machine=None, sent_count=None):
        """
        Args:
            state_machine: state machine that triggered the event
            sent_count: number of packets sent at time of event creation.
        """
        # will work but please do this properly
        self.type = self.TIMER_EXPIRED
        self.state_machine = state_machine
        self.sent_count = sent_count


class EventSessionTimeout(Event):
    """User's session should be terminated."""

    def __init__(self, state_machine=None):
        """
        Args:
            state_machine: state machine that triggered the event
        """
        self.type = self.TIMER_EXPIRED
        self.state_machine = state_machine


class EventMessageReceived(Event):
    """Message (EAP) Received. Radius Message event is a child"""

    def __init__(self, message, port_id):
        """
        Args:
            message:
            port_id: id of switch port where message was received.
        """
        # will work but please do this properly
        self.type = self.MESSAGE_RECEIVED
        self.message = message
        self.port_id = port_id

    def __eq__(self, other):
        return self.type == other.type and \
               self.message == other.message and \
               self.port_id == other.port_id

    def __repr__(self):
        return "%s(\"%s\", \"%s\")" % (self.__class__.__name__, self.message, self.port_id)


class EventPreemptiveEAPResponseMessageReceived(EventMessageReceived):
    """Radius Message Received."""

    def __init__(self, message, port_id, preemptive_eap_id):
        """
        Args:
            message:
            port_id:
            preemptive_eap_id:
        """
        super().__init__(message, port_id)
        self.preemptive_eap_id = preemptive_eap_id


class EventPortStatusChange(Event):
    """Port status has changed (up/down)"""

    def __init__(self, port_status):
        """
        Args:
            port_status (bool): True if port is enabled, False otherwise.
        """
        self.port_status = port_status
        if port_status:
            self.type = self.PORT_ENABLED
        else:
            self.type = self.PORT_DISABLED


class EventRadiusMessageReceived(EventMessageReceived):
    """Radius Message Received."""

    def __init__(self, message, state, attributes=None):
        """
        Args:
            message:
            state: the RADIUS state attribute
        """
        super().__init__(message, None)
        self.state = state
        self.attributes = attributes


class EventShutdown(Event):
    """Shutdown has been signaled (is this even used?)"""

    def __init__(self):
        # will work but please do this properly
        self.type = self.SHUTDOWN



"""Unittests for eap_state_machine.FullEAPStateMachine"""
# pylint: disable=missing-docstring

import logging
import tempfile
import unittest
from queue import Queue

from chewie.eap import Eap
from chewie.event import EventMessageReceived, EventRadiusMessageReceived, EventPortStatusChange
from chewie.mac_address import MacAddress
from chewie.message_parser import EapolStartMessage, IdentityMessage, Md5ChallengeMessage, \
    SuccessMessage, LegacyNakMessage, TtlsMessage, FailureMessage, EapolLogoffMessage
from chewie.radius_attributes import State
from chewie.state_machines.eap_state_machine import FullEAPStateMachine

from helpers import FakeTimerScheduler


def check_counters(_func=None, *,
                   expected_auth_counter=0, expected_failure_counter=0, expected_logoff_counter=0):
    """Decorator to check the handlers have been called the
     correct number of times at the end of each test"""

    def decorator_check_counters(func):
        def wrapper(self):
            start_auth_counter = self.auth_counter
            start_failure_counter = self.failure_counter
            start_logoff_counter = self.logoff_counter
            ret = func(self)
            self.assertEqual(self.auth_counter,
                             start_auth_counter + expected_auth_counter)
            self.assertEqual(self.failure_counter,
                             start_failure_counter + expected_failure_counter)
            self.assertEqual(self.logoff_counter,
                             start_logoff_counter + expected_logoff_counter)
            return ret

        return wrapper

    if _func is None:
        return decorator_check_counters
    else:
        return decorator_check_counters(_func)


class FullStateMachineStartTestCase(unittest.TestCase):
    # TODO tests could be more thorough, and test that
    # the correct packet (type/content) has been put in its respective queue.
    # Would also be nice to check that the states are correctly transitioned through,
    # and not just the final resting spot. Not sure how to do that - maybe parse the log??

    PORT_ID_MAC = MacAddress.from_string("00:00:00:00:00:01")

    def setUp(self):
        logger = logging.getLogger()
        logger.level = logging.DEBUG
        self.log_file = tempfile.NamedTemporaryFile()
        logger.addHandler(logging.FileHandler(self.log_file.name))

        self.eap_output_queue = Queue()
        self.radius_output_queue = Queue()
        self.timer_scheduler = FakeTimerScheduler()
        self.src_mac = MacAddress.from_string("00:12:34:56:78:90")
        log_prefix = "chewie.SM - port: %s, client: %s" % (self.src_mac, self.PORT_ID_MAC)

        self.sm = FullEAPStateMachine(self.eap_output_queue, self.radius_output_queue, self.src_mac,
                                      self.timer_scheduler,
                                      self.auth_handler, self.failure_handler, self.logoff_handler,
                                      log_prefix)
        # find ways to inject these - overriding consts isn't ideal
        self.MAX_RETRANSMITS = 3
        self.sm.MAX_RETRANS = self.MAX_RETRANSMITS
        self.sm.DEFAULT_TIMEOUT = 0.1
        self.sm.port_enabled = True
        self.sm.eap_restart = True

        self.auth_counter = 0
        self.failure_counter = 0
        self.logoff_counter = 0

    def tearDown(self):
        with open(self.log_file.name) as log:
            self.assertNotIn('aaaEapResp is true. but data is false. This should never happen',
                             log.read())

    def auth_handler(self, client_mac, port_id_mac, timer, vlan_name,
                     filter_id):  # pylint: disable=unused-argument
        self.auth_counter += 1
        print('Successful auth from MAC %s' % str(client_mac))

    def failure_handler(self, client_mac, port_id_mac):  # pylint: disable=unused-argument
        self.failure_counter += 1
        print('failure from MAC %s' % str(client_mac))

    def logoff_handler(self, client_mac, port_id_mac):  # pylint: disable=unused-argument
        self.logoff_counter += 1
        print('logoff from MAC %s' % str(client_mac))

    @check_counters
    def test_eap_start(self):
        # input EAPStart packet.
        # output EAPIdentityRequest on eap_output_q
        message = EapolStartMessage(self.src_mac)
        self.sm.event(EventMessageReceived(message, self.PORT_ID_MAC))
        self.assertEqual(self.sm.state, self.sm.IDLE)

        self.assertEqual(self.eap_output_queue.qsize(), 1)
        output = self.eap_output_queue.get_nowait()[0]
        self.assertIsInstance(output, IdentityMessage)
        self.assertEqual(self.radius_output_queue.qsize(), 0)

        return output  # Used by test_identity_response

    @check_counters
    def test_eap_restart(self):
        self.test_eap_start()
        message = EapolStartMessage(self.src_mac)
        self.sm.event(EventMessageReceived(message, self.PORT_ID_MAC))
        self.assertEqual(self.sm.state, self.sm.IDLE)

        self.assertEqual(self.eap_output_queue.qsize(), 1)
        output = self.eap_output_queue.get_nowait()[0]
        self.assertIsInstance(output, IdentityMessage)
        self.assertEqual(self.radius_output_queue.qsize(), 0)

    @check_counters
    def test_no_radius_state_attribute_after_restart(self):
        """The RADIUS State Attribute is returned by the RADIUS server,
        and MUST be sent back in the next RADIUS Message.
        Unless the authentication process has been restarted or its a new one.
        Here we restart part way through another."""
        self.test_md5_challenge_request()

        self.assertEqual(self.sm.radius_state_attribute.data_type.data(), b'random state')

        self.test_eap_restart()
        self.assertIsNone(self.sm.radius_state_attribute)

    @check_counters
    def test_timeout_failure_from_max_retransmits(self):
        """Go to timeout failure from exceeding max retransmits (also tests retransmitting)"""
        output0 = self.test_eap_start()

        old_radius_count = self.radius_output_queue.qsize()
        self.timer_scheduler.run_jobs()

        output1 = self.eap_output_queue.get_nowait()[0]
        output2 = self.eap_output_queue.get_nowait()[0]
        output3 = self.eap_output_queue.get_nowait()[0]
        self.assertIsInstance(output0, IdentityMessage)
        self.assertIsInstance(output1, IdentityMessage)
        self.assertEqual(output0, output1)
        self.assertEqual(output0, output2)
        self.assertEqual(output0, output3)
        self.assertEqual(self.radius_output_queue.qsize(), 0)

        self.assertEqual(self.sm.state, self.sm.TIMEOUT_FAILURE)
        self.assertEqual(self.eap_output_queue.qsize(), 0)
        self.assertEqual(old_radius_count, self.radius_output_queue.qsize())

    @check_counters(expected_auth_counter=1)
    def test_auth_success_after_timeout_failure2_from_max_retransmits(self):
        self.test_timeout_failure2_from_max_retransmits()
        self.eap_output_queue.queue.clear()
        self.test_success2()

    @check_counters(expected_auth_counter=1)
    def test_auth_success_after_timeout_failure2_from_aaa_timeout(self):
        self.test_timeout_failure2_from_aaa_timeout()
        self.test_success2()

    @check_counters(expected_auth_counter=1)
    def test_auth_success_after_timeout_failure_after_max_retransmits(self):
        self.test_timeout_failure_from_max_retransmits()
        self.test_success2()

    @check_counters()
    def test_leave_timeout_failure2_with_identiy_response(self):
        self.test_timeout_failure2_from_max_retransmits()
        start_eap_q_size = self.eap_output_queue.qsize()
        start_radius_q_size = self.radius_output_queue.qsize()
        message = IdentityMessage(self.src_mac, 25, Eap.RESPONSE, "host1user")
        self.sm.event(EventMessageReceived(message, self.PORT_ID_MAC))
        self.assertEqual(self.sm.state, self.sm.AAA_IDLE)

        self.assertEqual(self.eap_output_queue.qsize(), start_eap_q_size + 0)
        self.assertEqual(self.radius_output_queue.qsize(), start_radius_q_size + 1)
        self.assertIsInstance(self.radius_output_queue.get_nowait()[0], IdentityMessage)

    @check_counters
    def test_timeout_failure2_from_aaa_timeout(self):
        """no response from AAA server equals timeout_failure2"""
        self.test_md5_challenge_response()
        old_eap_count = self.eap_output_queue.qsize()
        old_radius_count = self.radius_output_queue.qsize()

        self.timer_scheduler.run_jobs()

        self.assertEqual(self.sm.state, self.sm.TIMEOUT_FAILURE2)
        self.assertEqual(old_eap_count, self.eap_output_queue.qsize())
        self.assertEqual(old_radius_count, self.radius_output_queue.qsize())

    @check_counters
    def test_timeout_failure2_from_max_retransmits(self):
        """If client does not respond when in passthrough mode,
         send again and again until max retransmit counter is reached."""
        self.test_md5_challenge_request()
        self.assertEqual(self.eap_output_queue.qsize(), 0)
        old_radius_count = self.radius_output_queue.qsize()

        self.timer_scheduler.run_jobs()

        self.assertEqual(self.sm.state, self.sm.TIMEOUT_FAILURE2)
        self.assertEqual(self.MAX_RETRANSMITS, self.eap_output_queue.qsize())
        self.assertEqual(old_radius_count, self.radius_output_queue.qsize())

    @check_counters(expected_auth_counter=1)
    def test_disabled_state(self):
        """move to disabled and then from disabled"""
        self.test_success2()
        # move to disabled. e.g. link down.
        self.sm.event(EventPortStatusChange(False))
        self.assertEqual(self.sm.state, self.sm.DISABLED)

        self.assertEqual(self.eap_output_queue.qsize(), 0)
        self.assertEqual(self.radius_output_queue.qsize(), 0)
        self.assertEqual(self.auth_counter, 1)
        # don't transition to initialize (still not enabled)
        message = EapolStartMessage(self.src_mac)
        self.sm.event(EventMessageReceived(message, self.PORT_ID_MAC))

        self.assertEqual(self.sm.state, self.sm.DISABLED)
        self.assertEqual(self.eap_output_queue.qsize(), 0)
        self.assertEqual(self.radius_output_queue.qsize(), 0)

        # port is enabled again
        self.sm.event(EventPortStatusChange(True))

        self.assertEqual(self.sm.state, self.sm.IDLE)

        self.assertEqual(self.eap_output_queue.qsize(), 1)
        self.assertEqual(self.radius_output_queue.qsize(), 0)

    @check_counters
    def test_identity_response(self):
        _id = self.test_eap_start().message_id
        # input EapIdentityResponse
        # output EapIdentityResponse on radius_output_q
        message = IdentityMessage(self.src_mac, _id, Eap.RESPONSE, "host1user")
        self.sm.event(EventMessageReceived(message, self.PORT_ID_MAC))

        self.assertEqual(self.sm.state, self.sm.AAA_IDLE)

        self.assertEqual(self.eap_output_queue.qsize(), 0)
        self.assertEqual(self.radius_output_queue.qsize(), 1)
        self.assertIsInstance(self.radius_output_queue.get_nowait()[0], IdentityMessage)

    @check_counters
    def test_md5_challenge_request(self):
        self.test_identity_response()

        eap_message = Md5ChallengeMessage(self.src_mac, 2, Eap.REQUEST,
                                          bytes.fromhex("74d3db089b727d9cc5774599e4a32a29"),
                                          b"host1user")
        self.sm.event(EventRadiusMessageReceived(eap_message, State.create(b"random state")))

        self.assertEqual(self.sm.state, self.sm.IDLE2)

        self.assertEqual(self.eap_output_queue.qsize(), 1)
        output = self.eap_output_queue.get_nowait()[0]
        self.assertIsInstance(output, Md5ChallengeMessage)

        self.assertEqual(self.radius_output_queue.qsize(), 0)
        return output

    @check_counters
    def test_md5_challenge_response(self):
        self.test_md5_challenge_request()

        message = Md5ChallengeMessage(self.src_mac, 2, Eap.RESPONSE,
                                      bytes.fromhex("3a535f0ee8c6b34fe714aa7dad9a0e15"),
                                      b"host1user")
        self.sm.event(EventMessageReceived(message, self.PORT_ID_MAC))

        self.assertEqual(self.sm.state, self.sm.AAA_IDLE)
        self.assertEqual(self.eap_output_queue.qsize(), 0)
        self.assertEqual(self.radius_output_queue.qsize(), 1)
        self.assertIsInstance(self.radius_output_queue.get_nowait()[0], Md5ChallengeMessage)

    @check_counters(expected_auth_counter=1)
    def test_success2(self):
        self.test_md5_challenge_response()

        message = SuccessMessage(self.src_mac, 3)
        self.sm.event(EventRadiusMessageReceived(message, None))

        self.assertEqual(self.sm.state, self.sm.SUCCESS2)
        self.assertEqual(self.eap_output_queue.qsize(), 1)
        self.assertIsInstance(self.eap_output_queue.get_nowait()[0], SuccessMessage)
        self.assertEqual(self.radius_output_queue.qsize(), 0)

    @check_counters(expected_auth_counter=2)
    def test_two_success2(self):
        self.test_success2()

        expiry_job = self.sm.session_timeout_job
        self.assertFalse(expiry_job.cancelled())

        self.test_success2()
        self.assertTrue(expiry_job.cancelled())

        expiry_job = self.sm.session_timeout_job
        self.assertFalse(expiry_job.cancelled())

    @check_counters(expected_auth_counter=2, expected_logoff_counter=1)
    def test_logoff2(self):
        """Test logoff from success2 state."""
        self.test_success2()

        # test the second logon expires the firsts session timeout event.
        expiry_job = self.sm.session_timeout_job
        self.assertFalse(expiry_job.cancelled())

        message = EapolLogoffMessage(self.src_mac)
        self.sm.event(EventRadiusMessageReceived(message, self.PORT_ID_MAC))

        self.assertTrue(expiry_job.cancelled())

        self.assertEqual(self.sm.state, self.sm.LOGOFF2)
        self.assertEqual(self.eap_output_queue.qsize(), 0)
        self.assertEqual(self.radius_output_queue.qsize(), 0)

        self.test_success2()

    @check_counters
    def test_logoff_from_idle2(self):
        """Test logoff from middle of authentication. should be ignored"""
        self.test_md5_challenge_request()

        message = EapolLogoffMessage(self.src_mac)
        self.sm.event(EventRadiusMessageReceived(message, self.PORT_ID_MAC))

        # should be in same state as when test_md5_challenge_request returned.

        self.assertEqual(self.sm.state, self.sm.IDLE2)
        self.assertEqual(self.eap_output_queue.qsize(), 0)
        self.assertEqual(self.radius_output_queue.qsize(), 0)

    @check_counters(expected_failure_counter=1)
    def test_failure2(self):
        self.test_md5_challenge_response()
        message = FailureMessage(self.src_mac, 3)
        self.sm.event(EventRadiusMessageReceived(message, None))

        self.assertEqual(self.sm.state, self.sm.FAILURE2)
        self.assertEqual(self.eap_output_queue.qsize(), 1)
        self.assertIsInstance(self.eap_output_queue.get_nowait()[0], FailureMessage)
        self.assertEqual(self.radius_output_queue.qsize(), 0)

    @check_counters
    def test_discard2(self):
        request = self.test_md5_challenge_request()

        message = Md5ChallengeMessage(self.src_mac, request.message_id + 10, Eap.RESPONSE,
                                      bytes.fromhex("3a535f0ee8c6b34fe714aa7dad9a0e15"),
                                      b"host1user")
        self.sm.event(EventMessageReceived(message, self.PORT_ID_MAC))
        self.assertEqual(self.sm.state, self.sm.IDLE2)
        self.assertEqual(self.eap_output_queue.qsize(), 0)
        self.assertEqual(self.radius_output_queue.qsize(), 0)

    @check_counters
    def test_discard(self):
        message = self.test_eap_start()
        # Make a message that will be discarded (id here is not sequential)
        message = IdentityMessage(self.src_mac, message.message_id + 10, Eap.RESPONSE, "host1user")
        self.sm.event(EventMessageReceived(message, self.PORT_ID_MAC))
        self.assertEqual(self.sm.state, self.sm.IDLE)

        self.assertEqual(self.eap_output_queue.qsize(), 0)
        self.assertEqual(self.radius_output_queue.qsize(), 0)

    @check_counters
    def test_ttls_request(self):
        self.test_md5_challenge_request()
        self.assertEqual(self.eap_output_queue.qsize(), 0)
        self.assertEqual(self.radius_output_queue.qsize(), 0)

        message = LegacyNakMessage(self.src_mac, 2, Eap.RESPONSE, 21)
        self.sm.event(EventMessageReceived(message, self.PORT_ID_MAC))
        self.assertEqual(self.sm.state, self.sm.AAA_IDLE)
        self.assertEqual(self.eap_output_queue.qsize(), 0)
        self.assertEqual(self.radius_output_queue.qsize(), 1)

        message = TtlsMessage(self.src_mac, 3, Eap.REQUEST, 0x20, b'')
        self.sm.event(EventRadiusMessageReceived(message, State.create(b"more random state")))
        self.assertEqual(self.sm.state, self.sm.IDLE2)
        self.assertEqual(self.eap_output_queue.qsize(), 1)
        self.assertEqual(self.radius_output_queue.qsize(), 1)

        message = TtlsMessage(self.src_mac, 3, Eap.RESPONSE, 0x00,
                              bytes.fromhex(
                                  '16030101280100012403032c36dbf8ee16b94b28efdb8c5603e07823f9b716557b5ef2624b026daea115760000aac030c02cc028c024c014c00a00a500a300a1009f006b006a0069006800390038003700360088008700860085c032c02ec02ac026c00fc005009d003d00350084c02fc02bc027c023c013c00900a400a200a0009e00670040003f003e0033003200310030009a0099009800970045004400430042c031c02dc029c025c00ec004009c003c002f00960041c011c007c00cc00200050004c012c008001600130010000dc00dc003000a00ff01000051000b000403000102000a001c001a00170019001c001b0018001a0016000e000d000b000c0009000a000d0020001e060106020603050105020503040104020403030103020303020102020203000f000101'))
        self.sm.event(EventMessageReceived(message, self.PORT_ID_MAC))
        self.assertEqual(self.sm.state, self.sm.AAA_IDLE)
        self.assertEqual(self.eap_output_queue.qsize(), 1)
        self.assertEqual(self.radius_output_queue.qsize(), 2)

        message = TtlsMessage(self.src_mac, 4, Eap.REQUEST, 0x00, bytes.fromhex(
            ' '))
        self.sm.event(EventRadiusMessageReceived(message, State.create(b"some more random state")))
        self.assertEqual(self.sm.state, self.sm.IDLE2)
        self.assertEqual(self.eap_output_queue.qsize(), 2)
        self.assertEqual(self.radius_output_queue.qsize(), 2)

    @check_counters(expected_auth_counter=1, expected_logoff_counter=1)
    def test_deauth_timer(self):
        self.sm.DEFAULT_SESSION_TIMEOUT = 2
        self.test_success2()

        self.timer_scheduler.run_jobs()
        self.assertEqual(self.sm.state, self.sm.LOGOFF2)

    @check_counters(expected_auth_counter=2, expected_logoff_counter=0)
    def test_port_flap(self):
        """Test logoff from success2 state."""
        self.test_success2()
        # Put port down
        self.sm.event(EventPortStatusChange(False))
        self.sm.event(EventPortStatusChange(True))

        self.assertEqual(self.sm.state, self.sm.NO_STATE)
        self.assertFalse(self.sm.aaa_success)
        self.assertFalse(self.sm.eap_success)

        self.assertEqual(self.eap_output_queue.qsize(), 0)
        self.assertEqual(self.radius_output_queue.qsize(), 0)

        self.test_success2()
